#!/usr/bin/python3
#
# Copyright [2023] Open Text.
#
# The only warranties for products and services of Open Text and its affiliates and licensors (“Open Text”) 
# are as may be set forth in the express warranty statements accompanying such products and services. 
# Nothing herein should be construed as constituting an additional warranty. Open Text shall not be liable 
# for technical or editorial errors or omissions contained herein. The information contained herein is 
# subject to change without notice.
#

# OES certificate class file

# This dataclass can be used in python 3.7 or higher
# from dataclasses import dataclass

# @dataclass
# class Services:
#     serviceName: str
#     certPath: str

import json
from typing import List

# OES service class
class OESServices(object):
    def __init__(self, serviceName:str, certPath:str):
        self.serviceName = serviceName
        self.certPath = certPath

    def __str__(self):
        return f"{self.serviceName}, {self.certPath}"
    
    
# OES certificate class
class OESCertificate(object):
    def __init__(self, subject:str, issuer:str, startDate:str, endDate:str, daysToExpiry:str, certType:str, fingerPrint:str, algorithm:str, services: List[OESServices]):
        self.subject = subject
        self.issuer = issuer
        self.startDate = startDate
        self.endDate = endDate
        self.daysToExpiry = daysToExpiry
        self.certType = certType
        self.fingerprint = fingerPrint
        self.algorithm = algorithm
        self.services = services

    def __str__(self):
        return f"{self.subject}, {self.issuer}, {self.startDate}, {self.endDate}, {self.daysToExpiry}, {self.certType}, {self.fingerprint}, {self.algorithm}"
   
#OES Service with certificate details
class OESServiceAndCert(object):
    def __init__(self, serviceName: str, certPath:str, subject:str, issuer:str, startDate:str, endDate:str, daysToExpiry:str, certType:str, fingerPrint:str, algorithm:str):
        self.serviceName = serviceName
        self.certPath = certPath
        self.subject = subject
        self.issuer = issuer
        self.startDate = startDate
        self.endDate = endDate
        self.daysToExpiry = daysToExpiry
        self.certType = certType
        self.fingerprint = fingerPrint
        self.algorithm = algorithm

    def __str__(self):
        return f"{self.serviceName}, {self.certPath}, {self.subject}, {self.issuer}, {self.startDate}, {self.endDate}, {self.daysToExpiry}, {self.certType}, {self.fingerprint}, {self.algorithm}"
