#!/usr/bin/python3
#
# Copyright [2023] Open Text.
#
# The only warranties for products and services of Open Text and its affiliates and licensors (“Open Text”) 
# are as may be set forth in the express warranty statements accompanying such products and services. 
# Nothing herein should be construed as constituting an additional warranty. Open Text shall not be liable 
# for technical or editorial errors or omissions contained herein. The information contained herein is 
# subject to change without notice.
#

import sys
import subprocess
import os
from subprocess import call
import json
import datetime
from ast import literal_eval
import shlex
import os.path
import logging
import csv
from typing import List
import oes_cert_mgmt_utils
from oes_cert_mgmt_certificate import OESCertificate, OESServices, OESServiceAndCert

#List of certificates and their services
OESCertificateArray = []

#List of services with their certificate details
OESServiceAndCertArray = []

#Failed to process any certificate?
failedcerts = 0

now = datetime.datetime.now()
dateFormater=(now.strftime("%b %d %H:%M:%S %Y"))
current_time = datetime.datetime.strptime(dateFormater, '%b %d %H:%M:%S %Y')

#json files for dumping certificates data
jsonsrvFilePath = oes_cert_mgmt_utils.VarDataPath + "/certlist-service.json"
jsonsrvFilePathBak1 = oes_cert_mgmt_utils.VarDataPath + "/certlist-service.json.bak.1"
jsonsrvFilePathBak2 = oes_cert_mgmt_utils.VarDataPath + "/certlist-service.json.bak.2"
jsoncertFilePath = oes_cert_mgmt_utils.VarDataPath + "/certlist-cert.json"
jsoncertFilePathBak1 = oes_cert_mgmt_utils.VarDataPath + "/certlist-cert.json.bak.1"
jsoncertFilePathBak2 = oes_cert_mgmt_utils.VarDataPath + "/certlist-cert.json.bak.2"
CSVFilePath = oes_cert_mgmt_utils.VarDataPath + "/jsonoutput.csv"

ServiceBinPath = oes_cert_mgmt_utils.BinaryPath + "/"
ServiceCertsPath = oes_cert_mgmt_utils.ServiceCertsPath + "/"
CertListwithServices=[]
ServiceListWithPath=[]

#While getting certificates to distinguish to get services using eDirectory server certificate
#using subject and CA
edirsubcahash = False

#To store hash of eDirectory server certificates subject and hash
eDirCertSubHash = ""
eDirCertCAHash = ""

#Get subject and CA hash eDirectory server certificates
#Subject and CA for both RSA and EC certficates are same
#Getting Suject and CA Hash for one certificate is same as for other
def getedircertsubandcahash():
    edircertpath = ""

    if os.path.exists(oes_cert_mgmt_utils.eDirRSAPath):
        edircertpath = oes_cert_mgmt_utils.eDirRSAPath
    elif os.path.exists(oes_cert_mgmt_utils.eDirECPath):
        edircertpath = oes_cert_mgmt_utils.eDirECPath

    if not edircertpath:
        log.error("eDirectory certificates are not found")
        return 1

    global eDirCertSubHash, eDirCertCAHash
    cmd = "openssl x509 -subject_hash -noout -in " + edircertpath
    output = oes_cert_mgmt_utils.run_command(cmd)
    if(output == 0):
        logging.error("Could not get subject hash of the certificate: " + str(edircertpath))
        return 1
    eDirCertSubHash  = (str(output).replace("\n", ""))

    cmd = "openssl x509 -issuer_hash -noout -in " + edircertpath
    output = oes_cert_mgmt_utils.run_command(cmd)
    if(output == 0):
        logging.error("Could not get Issuer hash of the certificate: " + str(edircertpath))
        return 1
    eDirCertCAHash  = (str(output).replace("\n", ""))

    logging.info("eDirectory certificates subject hash: " + str(eDirCertSubHash))
    logging.info("eDirectory certificates CA hash: " + str(eDirCertCAHash))

    #Just to ensure that EC Cert is not corrupted, check for subject hash
    if(edircertpath == oes_cert_mgmt_utils.eDirRSAPath):
        cmd = "openssl x509 -subject_hash -noout -in " + oes_cert_mgmt_utils.eDirECPath
        output = oes_cert_mgmt_utils.run_command(cmd)
        if(output == 0):
            logging.error("Could not get subject hash of the certificate: " + str(edircertpath))
            return 1

    return 0

#Service Name, Certificate Path, Subject Name, Start Date, End Date, CA Signed Or Self Signed.
def convertJsontoCSV():
    with open(jsonsrvFilePath) as json_file:
        jsondata = json.load(json_file)

 
    data_file = open(CSVFielPath, 'w', newline='')
    csv_writer = csv.writer(data_file)

    count = 0
    for data in jsondata:
        if count == 0:
            header = data.keys()
            csv_writer.writerow(header)
            count += 1
        csv_writer.writerow(data.values())

    data_file.close()


def Write_Json_Object_to_file_rest(val):
    with open(jsonFilePath, 'w', encoding='utf-8') as jsonf: 
        if val == 1:
            print("Certlist")
            jsonString = json.dumps(CertListwithServices, indent=5)
        else:
            print("Servicelist")
            jsonString = json.dumps(ServiceListWithPath, indent=5)
        jsonf.write(jsonString)
    return jsonString

# Func to write JsonObject to file
# Keep two back up files as bak.1 and bak.2
def Write_Json_Object_to_file(listmode):
    if listmode == oes_cert_mgmt_utils.SERVICE:
        jsonFilePath = jsonsrvFilePath
        file_exists = os.path.exists(jsonsrvFilePath)
        if file_exists:
            file_bak_exists = os.path.exists(jsonsrvFilePathBak1)
            if file_bak_exists:
                os.rename(jsonsrvFilePathBak1, jsonsrvFilePathBak2)
            os.rename(jsonsrvFilePath, jsonsrvFilePathBak1)
    else:
        jsonFilePath = jsoncertFilePath
        file_exists = os.path.exists(jsonFilePath)
        if file_exists:
            file_bak_exists = os.path.exists(jsoncertFilePathBak1)
            if file_bak_exists:
                os.rename(jsoncertFilePathBak1, jsoncertFilePathBak2)
            os.rename(jsoncertFilePath, jsoncertFilePathBak1)

    # wrtie logic to delete older file
    if (listmode == oes_cert_mgmt_utils.CERTIFICATE):
        json_data=json.dumps(OESCertificateArray, default=lambda o: o.__dict__, indent=4)
    else:
        json_data=json.dumps(OESServiceAndCertArray, default=lambda o: o.__dict__, indent=4) 
    with open(jsonFilePath, 'w', encoding='utf-8') as jsonfile: 
        if listmode == oes_cert_mgmt_utils.SERVICE:
            jsonString = json.dumps(CertListwithServices, indent=5)
        else:
            jsonString = json.dumps(ServiceListWithPath, indent=5)
        jsonfile.write(json_data)
    
    print("Certificates in use along with service details are written to the file " + jsonFilePath);


def parse_jsonfile(filename):
    with open(filename) as jsonFile:
        jsonObject = json.load(jsonFile)
        jsonFile.close()
    service = jsonObject['servicename']
    servicecertpath = jsonObject['certpath']
    certpathtoprocess = jsonObject['certpathtoprocess']
    return service,servicecertpath,certpathtoprocess
        

def get_servicecert_details(servicescript, serviceName):
    global failedcerts
    ServiceJsonFile = oes_cert_mgmt_utils.ServiceCertsPath +'/' + serviceName.lower() + ".json"
    file_exists = os.path.exists(servicescript)
    if file_exists:
        logging.debug("Processing Service: " + str(serviceName) + " with service executable: " + str(servicescript))
        s2_out = subprocess.run(['python3', servicescript])
        serviceexitcode = s2_out.returncode
        logging.debug("Service executable " + str(servicescript) + " has returned " + str(serviceexitcode))
        
        #Service using non-eDirectory certificate
        #Certificate details are received from service in JSON file
        if serviceexitcode == oes_cert_mgmt_utils.NONEDIRCERT:
            file_exists = os.path.exists(ServiceJsonFile)
            if file_exists:
                service,servicecertpath,certificatetoprocess = parse_jsonfile(ServiceJsonFile)
                logging.debug("ServiceCertPath: " + str(servicecertpath) + " CertificateToProcess: " + str(certificatetoprocess))
                if (servicecertpath and certificatetoprocess):
                    if(os.path.exists(certificatetoprocess)):
                        certresult = Get_Certificate_Details(serviceName, servicecertpath, "NONE", certificatetoprocess)
                        if(certresult):
                            failedcerts = failedcerts + 1
                        if (servicecertpath != certificatetoprocess):
                            os.remove(certificatetoprocess)
                            logging.debug("Removed service certificate provided for processing " + str(certificatetoprocess))
                    else:
                        logging.error("Service certificate file to process " + str(certificatetoprocess) + " is not found")
                else:
                    logging.error("Service " + str(serviceName) + " Certificate details are not available from json file")
                os.remove(ServiceJsonFile)
                logging.debug("Removed service json file: " + str(ServiceJsonFile))
            else:
                logging.info(str(serviceName) + " JSON file: " + str(ServiceJsonFile) + " not found")
        #Service is using eDirectory RSA Certificate
        elif serviceexitcode == oes_cert_mgmt_utils.EDIRRSACERT:
            edircertpath=oes_cert_mgmt_utils.eDirRSAPath
            file_exists = os.path.exists(edircertpath)
            if file_exists:
                    certresult = Get_Certificate_Details(serviceName, edircertpath, edircertpath, edircertpath)
                    if(certresult):
                        failedcerts = failedcerts + 1
            else:
                    logging.error("Service Certificate path or eDirectory RSA Certificate path is NULL")
        #Service is using eDirectory EC Certificate
        elif serviceexitcode == oes_cert_mgmt_utils.EDIRECCERT:
            edircertpath=oes_cert_mgmt_utils.eDirECPath
            file_exists = os.path.exists(edircertpath)
            if file_exists:
                #ervicecertpath=parse_jsonfile(ServiceJsonFile)
                #f servicecertpath and edircertpath:
                    certresult = Get_Certificate_Details(serviceName, edircertpath, edircertpath, edircertpath)
                    if(certresult):
                        failedcerts = failedcerts + 1
            else:
                    logging.error("Service Certificate path or eDirectory ECDSA certificate path is NULL")
        elif serviceexitcode == oes_cert_mgmt_utils.NOCERT:
            logging.error("Service is not configured to use any certificate")
        elif serviceexitcode == oes_cert_mgmt_utils.FAIL:
            logging.error("Failed to get certificate details of the service")
        else:
            logging.error( "Invalid return value from service: " + str(serviceexitcode))

    else:
        logging.info("Executable for service " + str(servicescript) + " doesn't exist")

#Compare two strings
def compare_strings(str1, str2):
    count1 = 0
    count2 = 0
      
    for i in range(len(str1)):
        if str1[i] >= "0" and str1[i] <= "9":
            count1 += 1
      
    for i in range(len(str2)):
        if str2[i] >= "0" and str2[i] <= "9":
            count2 += 1
      
    return count1 == count2
#end

#API to Get Certificate_Details using openssl command
def Get_Certificate_Details(ServiceName, servicecertpath, edircertpath, certificateToProcess):
    global edirsubcahash
    certPathDisplay = ""
    if edircertpath != "NONE":
        Certpathstr = edircertpath
        certPathDisplay = edircertpath
    else:
        Certpathstr = certificateToProcess
        certPathDisplay = servicecertpath
    Path = "Certificate Path:" + Certpathstr
    CertExists = os.path.exists(Certpathstr)
    if CertExists == 0:
        logging.error("Service Certificate: " + Certpathstr + "doesn't exist");
        return
    
    cmd = "openssl x509 -noout -fingerprint -sha256 -inform pem -in " + Certpathstr
    output = oes_cert_mgmt_utils.run_command(cmd)
    if(output == 0):
        logging.error("Could not get certificate details for the certificate: " + str(Certpathstr))
        return 1
    string =(output.replace("\n", ""))
    fingerprint = string.split('=')[1]

    #If certificate is already there in the list, add to service list of certificate
    #Add to certificate list if fingeprint matches and not subject and ca of eDir certificate
    if (edirsubcahash == False):
        for i in range(0, len(OESCertificateArray)):
            if(OESCertificateArray[i].fingerprint == fingerprint):
                service=OESServices(ServiceName, certPathDisplay)
                OESCertificateArray[i].services.append(service)
                return

    cmd = "openssl x509 -startdate -noout -in " + Certpathstr 
    output = oes_cert_mgmt_utils.run_command(cmd)
    string =(output.replace("\n", ""))
    startDate = string.split('=')[1]

    cmd = "openssl x509 -enddate -noout -in " + Certpathstr 
    output =oes_cert_mgmt_utils.run_command(cmd)
    string =(output.replace("\n", ""))
    endDate = string.split('=')[1]
    
    cmd = "openssl x509 -subject -noout -in " + Certpathstr 
    output = oes_cert_mgmt_utils.run_command(cmd)
    subject=(output.replace("\n", ""))
    if "subject=" in subject:
        subject = subject.replace("subject=", "")
   
    cmd = "openssl x509 -issuer -noout -in " + Certpathstr 
    output = oes_cert_mgmt_utils.run_command(cmd)   
    issuer=(output.replace("\n", ""))
    if "issuer=" in issuer:
        issuer = issuer.replace("issuer=", "")

    cmd = "openssl x509 -subject_hash -noout -in " + Certpathstr
    output = oes_cert_mgmt_utils.run_command(cmd)
    subjecthash = (str(output).replace("\n", ""))

    cmd = "openssl x509 -issuer_hash -noout -in " + Certpathstr
    output = oes_cert_mgmt_utils.run_command(cmd)
    issuerhash = (str(output).replace("\n", ""))

    logging.debug("Subject hash: " + str(subjecthash) + " Issuer hash: " + str(issuerhash))

    if (subjecthash == issuerhash):
        Signed = "self-signed"
    else:
        Signed = "CA-signed"
    certType =Signed
    
    cmd = "openssl x509 -text -noout -in " + Certpathstr + " | grep \"Public Key Algorithm\"" 
    output = subprocess.check_output(cmd, shell=True, universal_newlines=True)
    string =(output.replace("\n", ""))
    algorithm = string.split(':')[1]
    
    preTempCertExpiryTime = endDate.replace("GMT", "")
    tempCertExpiryTime=(preTempCertExpiryTime.replace("\n", ""))
    certExpiryTime=(tempCertExpiryTime.rstrip())

    now = datetime.datetime.now()
    currTime=(now.strftime("%b %d %H:%M:%S %Y"))

    certTime = datetime.datetime.strptime(certExpiryTime, '%b %d %H:%M:%S %Y')
    currentTime = datetime.datetime.strptime(currTime, '%b %d %H:%M:%S %Y')
    time_remaining = certTime - currentTime

    days = time_remaining.days
    if (days < 0):
        daysToExpiry = "0"
    else:
        daysToExpiry = str(days)

    #Looking for all the certificates
    if (edirsubcahash == False):
        service=OESServices(ServiceName, certPathDisplay)
        certificate = OESCertificate(subject, issuer, startDate, endDate, daysToExpiry, certType, fingerprint, algorithm, services=[service])
        OESCertificateArray.append(certificate)
    else:   #Looking for the services using eDirectory certificate only, match using subject and ca hash
        if(subjecthash == eDirCertSubHash and issuerhash == eDirCertCAHash):
            if not len(OESCertificateArray):
                service=OESServices(ServiceName, certPathDisplay)
                certificate = OESCertificate(subject, issuer, startDate, endDate, daysToExpiry, certType, fingerprint, algorithm, services=[service])
                OESCertificateArray.append(certificate)
            else:
                service=OESServices(ServiceName, certPathDisplay)
                OESCertificateArray[0].services.append(service)

    return 0

def edir_get_rsa_certificate_details():
    global failedcerts
    file_exists = os.path.exists(oes_cert_mgmt_utils.eDirRSAPath)
    if file_exists:
        certresult = Get_Certificate_Details("eDirectory RSA", oes_cert_mgmt_utils.eDirRSAPath,oes_cert_mgmt_utils.eDirRSAPath, oes_cert_mgmt_utils.eDirRSAPath)
        if(certresult):
            failedcerts = failedcerts + 1
    else:
        logging.error('RSA Certificate of eDirectory server is not found')

def edir_get_ecdsa_certificate_details():
    global failedcerts
    file_exists = os.path.exists(oes_cert_mgmt_utils.eDirECPath)
    if file_exists:
        certresult = Get_Certificate_Details("eDirectory ECDSA",oes_cert_mgmt_utils.eDirECPath,oes_cert_mgmt_utils.eDirECPath, oes_cert_mgmt_utils.eDirECPath)
        if (certresult):
            failedcerts = failedcerts + 1
    else:
         logging.error('ECDSA Certificate of eDirectory server is not found')


#Get list of all certificates with the consumers(services) details
def OES_get_certificate_details():
    logging.info("Retrieving certificate details of services")
    
    edir_get_rsa_certificate_details()
    edir_get_ecdsa_certificate_details()

    for ServiceName in oes_cert_mgmt_utils.OES_Services:
        ServiceGetCertPath = oes_cert_mgmt_utils.get_service_exec_path(ServiceName, oes_cert_mgmt_utils.GETCERTBINARY)
        if len(ServiceGetCertPath) != 0:
            get_servicecert_details(ServiceGetCertPath, ServiceName);
    
    # print(OESCertificateArray)
    if(failedcerts):
        print("Failed to process one or more certificates. Refer to log file " + str(oes_cert_mgmt_utils.CertificateLogFile) + " for more details")
    return OESCertificateArray

#Get list of individual services with their certificate details
def OES_get_serviceandcert_details():
    global edirsubcahash
    logging.info("Retrieving details of services")
    edirsubcahash = False
    OESCertList = OES_get_certificate_details()
   
    for oescert in OESCertList:
        for oesservice in oescert.services:
            OESServiceDetails = OESServiceAndCert(oesservice.serviceName, oesservice.certPath, oescert.subject, oescert.issuer, oescert.startDate, oescert.endDate, oescert.daysToExpiry, oescert.certType, oescert.fingerprint, oescert.algorithm)
            OESServiceAndCertArray.append(OESServiceDetails)

    return OESServiceAndCertArray

#Services using eDirectory certificate using eDirectory certificate subject and CA
#There will be only one certificate object in the list with services using the eDirectory certificate
#Used for reoconfiguration of services when eDirectory certificate is changed under /etc/ssl/servercerts/
#Fingerprint based matching of certificates won't work if services copy the certificate to their specific
#locations and format like JKS and DER
def OES_get_edir_cert_subca_services():
    global edirsubcahash
    logging.info("Retrieving services using certificates same as eDirectory certificate subject and CA")
    edirsubcahash = True
    gotsubcahash = getedircertsubandcahash()
    if(gotsubcahash):
        logging.error("Failed to get services using certificates same as eDirectory certificate subject and CA")
        return OESCertificateArray
    else:
        OESCertList = OES_get_certificate_details()
        return OESCertList
            
