#!/usr/bin/python3
#
# Copyright [2023] Open Text.
#
# The only warranties for products and services of Open Text and its affiliates and licensors (“Open Text”)
# are as may be set forth in the express warranty statements accompanying such products and services.
# Nothing herein should be construed as constituting an additional warranty. Open Text shall not be liable
# for technical or editorial errors or omissions contained herein. The information contained herein is
# subject to change without notice.
#


import subprocess
import os
from subprocess import call
import json
import sys
from ast import literal_eval
import shlex
import os.path
import logging
import shutil
import datetime
import oes_cert_mgmt_utils
import oes_cert_mgmt_listutils
# Reconfigure code Start
CertType = "0"

#Reconfig operations
CERTCHANGE = 1  #An existing certificate being replaced
RECONFIG = 2    #Admin wants to reconfigure set of services with a certificate
EDIRCERTCHANGE = 3  #Indicate services using eDirectory server certificate
MOVETOEDIRCERT = 4  #Move a set of services to use eDirectory server certificate

#code to create reconfigure list based on fingerprint from the service array
def search_srcFingerPrint(SrcFingerPrintstr, ServiceArray,FingerPrintArray):
    print(ServiceArray)
    ReconfigServices=[]
    SrcFingerPrint=(SrcFingerPrintstr.replace("\n", ""))
    fingerprint = SrcFingerPrint.split('=')
    for x in range(len(FingerPrintArray)):
        print(fingerprint[1])
        print('******************************')
        print(FingerPrintArray[x])
        if FingerPrintArray[x] == fingerprint[1]:
            ReconfigServices.append(ServiceArray[x].lower())
    return ReconfigServices


#code to create reconfigure list based on subjectname from the service array
def search_srcSubjectName(SrcSubjectNamestr,SrcFingerPrint,ServiceArray,SubjectNameArray,FingerPrintArray):
    ReconfigServices=[]
    SrcSubjectName=(SrcSubjectNamestr.replace("\n", ""))
    for x in range(len(SubjectNameArray)):
        if SubjectNameArray[x] == SrcSubjectName:
            ReconfigServices.append(ServiceArray[x])
    return ReconfigServices


def buildreconfigureserviceswith_samesubjectandissuer(SrcSubjectNamestr, srcissuerstr, ServiceArray,SubjectNameArray, IssuerArray):
    ReconfigServices=[]
    for x in range(len(SubjectNameArray)):
        SrcSubjectName=(SrcSubjectNamestr.replace("\n", ""))
        srcissuer=(srcissuerstr.replace("\n", ""))
        if SubjectNameArray[x] == SrcSubjectName and IssuerArray[x] == srcissuer:
            ReconfigServices.append(ServiceArray[x])

    return ReconfigServices

#code to get the fingerprint from certificate 
def get_fingerPrint(CertPath):
    cmd = "openssl x509 -noout -fingerprint -sha256 -inform pem -in " + str(CertPath)

    fingerprint = subprocess.check_output(cmd, shell=True, universal_newlines=True)
    fingerprintsplit = fingerprint.split('=')
    return fingerprintsplit[1].strip()

#code to get the subjectname from certificate 
def get_subjectName(CertPath):
    cmd = "openssl x509 -subject -noout -in " + CertPath
    subject = subprocess.check_output(cmd, shell=True, universal_newlines=True)
    return subject


#code to get the issurname from certificate 
def get_issuername(CertPath):
    cmd = "openssl x509 -issuer -noout -in " + CertPath
    issuer = subprocess.check_output(cmd, shell=True, universal_newlines=True)
    return issuer

#code to get the algorithm from certificate 
def get_algorithm(CertPath):
    cmd = "openssl x509 -text -noout -in " + Certpathstr + " | grep \"Public Key Algorithm\"" 
    output = subprocess.check_output(cmd, shell=True, universal_newlines=True)
    algorithm = output.split(':')
    return algorithm[1]

#Validate new certificate for expiry and self signed
def validate_newcert(certpath):
    cmd = "openssl x509 -subject_hash -noout -in " + certpath
    output = oes_cert_mgmt_utils.run_command(cmd)
    if(output == 0):
        logging.error("Could not get subject hash of the certificate: " + str(certpath))
        return 1
    subjecthash = (str(output).replace("\n", ""))

    cmd = "openssl x509 -issuer_hash -noout -in " + certpath
    output = oes_cert_mgmt_utils.run_command(cmd)
    if(output == 0):
        logging.error("Could not get issuer hash of the certificate: " + str(certpath))
        return 1
    issuerhash = (str(output).replace("\n", ""))

    logging.debug("Subject hash: " + str(subjecthash) + " Issuer hash: " + str(issuerhash))

    if (subjecthash == issuerhash):
        logging.error("Certificate is self signed certificate. Not proceeding with reconfiguration")
        return 1
    else:
        logging.debug("Certificate is a CA signed certificate")

    #Check for certificate expiry
    now = datetime.datetime.now()
    currTime = (now.strftime("%b %d %H:%M:%S %Y"))
    currTime = datetime.datetime.strptime(currTime, '%b %d %H:%M:%S %Y')
    logging.debug("Current Time: " + str(currTime))

    cmd = "openssl x509 -enddate -noout -in " + certpath
    output =oes_cert_mgmt_utils.run_command(cmd)
    if(output == 0):
        logging.error("Could not get expiration date of the certificate: " + str(certpath))
        return 1
    string =(output.replace("\n", ""))
    endDate = string.split('=')[1]

    preTempCertExpiryTime = endDate.replace("GMT", "")
    tempCertExpiryTime=(preTempCertExpiryTime.replace("\n", ""))
    certExpiryTime=(tempCertExpiryTime.rstrip())
    certExpiryTime = datetime.datetime.strptime(certExpiryTime, '%b %d %H:%M:%S %Y')
    logging.debug( "Certificate Expire Time: " + str(certExpiryTime))
    if (currTime >= certExpiryTime):
        logging.error("Certificate has expired. Not proceeding with reconfiguration")
        return 1
    else:
        logging.debug("Certificate is valid for reconfiguration")
        return 0


#Replace a certificate with new certificate
def reconfigure_replacecert(newcert, newcertkey, newcacert, currentcert, currentcertkey, restart, services):
    #Not allowed if certificate being replaced is eDirectory server certificate
    if(currentcert == oes_cert_mgmt_utils.eDirRSAPath or currentcert == oes_cert_mgmt_utils.eDirECPath):
        logging.error("Certificate being replaced is eDirectory Server certificate. Operation is not allowed")
        return 1

    #Compare finger prints of current and new certificates and compare them
    logging.debug("Getting finger print for the current certificate: " + str(currentcert))
    srcfingerprint = get_fingerPrint(currentcert)
    if not srcfingerprint:
        logging.error("Could not get finger print of current certificate: " + str(currentcert))
        return 1
    logging.debug("Finger print of the certificate: " + srcfingerprint)

    logging.debug("Getting finger print of the new certificate: " + str(newcert))
    dstfingerprint = get_fingerPrint(newcert)
    if not dstfingerprint:
        logging.error("Could not get finger print of certificate: " + str(newcert))
        return 1
    logging.debug("Finger print of the new certificate: " + dstfingerprint)

    if(srcfingerprint == dstfingerprint):
        logging.error("Current and new certificate are same. Nothing to reconfigure")
        return 0

    #Find the certificate matching the current certificate
    OESCertArray = oes_cert_mgmt_listutils.OES_get_certificate_details()
    foundcert = 0
    for OESCert in OESCertArray:
        logging.debug("Fingerprint of Certificate from Certificate array: " + OESCert.fingerprint)
        if(srcfingerprint == OESCert.fingerprint):
            logging.debug("Found the certificate with matching finger print: "  +  str(OESCert))
            foundcert = 1
            CertObject = OESCert
        else:
            continue
    if (not foundcert):
        logging.info("No service is found to be using the certificate: " + str(currentcert))
        logging.info("No need to reconfigure any service")
        return 0

    #Take backup of certficate being replaced and its private key file
    backupfilepath = currentcert + ".cert-mgmt-bak"
    shutil.copyfile(currentcert, backupfilepath)

    backupfilepath = currentcertkey + ".cert-mgmt-bak"
    shutil.copyfile(currentcertkey, backupfilepath)

    #Get the names of all services using the currentcert
    CertServices = []
    for CertService in CertObject.services:
        CertServices.append(CertService.serviceName)
    logging.info("List of services to reconfigure with certificate replacement: " + str(CertServices))
    
    #Reconfigure services with new certificate
    configresult = reconfigureservices_withcert(newcert, newcertkey, newcacert, restart, CertServices, CERTCHANGE)
    return configresult
#End of replacing a certificate with new certificate

def reconfigure_withcert(newcert, newcertkey, newcacert, restart, listofservices):
    #reconfig bot allowed with eDirectory server certificate
    if(newcert == oes_cert_mgmt_utils.eDirRSAPath or newcert == oes_cert_mgmt_utils.eDirECPath):
        logging.error("reconfig operaion with eDirectory server certificate is not allowed. Use movetoedircert for moving services to eDirectory server certificate")
        return 1
    configresult = reconfigureservices_withcert(newcert, newcertkey, newcacert, restart, listofservices, RECONFIG)
    return configresult

#Reconfigure services with a new certificate
def reconfigureservices_withcert(newcert, newcertkey, newcacert, restart, listofservices, operationcode):
    failcount = 0
    for service in listofservices:
        lowerservice = service.lower()
        logging.debug("Service name in lower case: " +  str(lowerservice))
        if (lowerservice.find("edirectory") != -1):
            logging.debug("Skipping eDirectory service")
            continue
        ServiceReconfExecPath = oes_cert_mgmt_utils.get_service_exec_path(service, oes_cert_mgmt_utils.RECONFIGBINARY)
        if(len(ServiceReconfExecPath) and os.path.exists(ServiceReconfExecPath)):
            logging.debug("Calling " + str(service) + " Reconfig executable " + str(ServiceReconfExecPath))
            reconfresult = subprocess.run(['python3', ServiceReconfExecPath, '--operation', str(operationcode), '--certificate', str(newcert), '--privatekey', str(newcertkey), '--cacertificate', str(newcacert), '--restart', str(restart)])
            if(not reconfresult.returncode):
                logging.info("Service " + str(service) + " is reconfitured to use certificate " + str(newcert))
            elif(reconfresult.returncode == oes_cert_mgmt_utils.OPUNSUPPORTED):
                logging.info("Service " + str(service) + " does not support reconfiguration")
            else:
                logging.error("Service " + str(service) + " has failed to reconfigure to use certificate " + str(newcert) + " with error code " + str(reconfresult.returncode))
                failcount = failcount + 1
        else:
            logging.debug("Executable to reconfigure service " + str(service) + " is not found")

    if(failcount):
        print("One or more services fail to reconfigure. Refer to log file for more details")
        logging.error("One or more services fail to reconfigure")
    else:
        print("Reconfiguration of services is completed")
        logging.info("Reconfiguration of services is completed")

    if(failcount):
        return 1
    else:
        return 0
#End of reconfiguring services with a new certificate


#eDirectory server certificate is changed
def reconfigure_edircert(restart):
    #Get list of services using eDirectory server certificates
    #Match the certificate using certificate subject and CA
    OESCertObjectList = oes_cert_mgmt_listutils.OES_get_edir_cert_subca_services()
    logging.debug("Number of objects in OESCertObjectList: " + str(len(OESCertObjectList)))
    if(len(OESCertObjectList) == 0):
        return 1
    eDirCertServices = []
    for OESCert in OESCertObjectList:
        logging.debug("Certificate Details " + str(OESCert))
        CertServices = OESCert.services
        for certservice in CertServices:
            if(certservice.serviceName != "eDirectory RSA" and certservice.serviceName != "eDirectory ECDSA"):
                logging.debug("Service name being added to eDirectory services list for reconfiguration " + str(certservice.serviceName))
                eDirCertServices.append(certservice)
    if(len(eDirCertServices)):
        logging.debug("Number of services using eDirectory server certificate: " + str(len(eDirCertServices)))
        failcount = 0
        for eDirCertService in eDirCertServices:
            ServiceReconfExecPath = oes_cert_mgmt_utils.get_service_exec_path(eDirCertService.serviceName, oes_cert_mgmt_utils.RECONFIGBINARY)
            if (len(ServiceReconfExecPath) and os.path.exists(ServiceReconfExecPath)):
                logging.debug("Service Reconfigure Executable path: " + ServiceReconfExecPath)
                reconfresult = subprocess.run(['python3', ServiceReconfExecPath, '--operation', str(3), '--restart', str(restart)])
                if(not reconfresult.returncode):
                    logging.info("Service successfully reconfigured with new eDirectory certificate")
                else:
                    logging.error("Service failed to reconfigure with new eDirectory certificate")
                    failcount = failcount + 1
            else:
                logging.error("Service Reconfiguration executale missing for the service: " + eDirCertService.serviceName)
    if(failcount):
        print("One or more services fail to reconfigure. Refer to log file for more details")
        logging.error("One or more services fail to reconfigure")
    else:
        print("All services are successfully reconfigured")
        logging.info("All services are successfully reconfigured")

    if(failcount):
        return 1
    else:
        return 0
#end of eDirectory server certificate change event handler


#Move services to eDirectory server certificate
def reconfigure_movetoedircert(servicestomove, restartService):
    failcount = 0
    for service in servicestomove:
        ServiceReconfExecPath = oes_cert_mgmt_utils.get_service_exec_path(service, oes_cert_mgmt_utils.RECONFIGBINARY)
        if (len(ServiceReconfExecPath) and os.path.exists(ServiceReconfExecPath)):
            logging.debug("Service Reconfiguration executable path getting executed " + ServiceReconfExecPath)
            reconfresult = subprocess.run(['python3', ServiceReconfExecPath, '--operation', str(4), '--restart', str(restartService)])  
            if(reconfresult.returncode):
                logging.error("Service " + str(service) + " Reconfiguration to move to eDirectory server certificate failed with error: " + str(reconfresult.returncode))
                failcount = failcount + 1
            else:
                #logging.info("Service " + servce + " successfully reconfigured to use eDirectory server certificate")
                logging.info("Service " + str(service) + " successfully reconfigured to use eDirectory server certificate")
        else:
            #logging.info("Service reconfiguration executable " + ServiceReconfExecPath + "missing")
            logging.error("Service reconfiguration executable " + ServiceReconfExecPath + "missing")

    if(failcount):
        print("One or more services fail to move to eDirctory server certificate")
        logging.error("One or more services fail to move to eDirectory server certificate")
    else:
        print("All selected services are successfully moved to eDirectory server certificate")

    if(failcount):
        return 1
    else:
        return 0
#End of moving services to eDirectory server certificate
    
